--  Z180_EC_Util_CPLD_v7.vhd
--
--  VHD for Peripheral Chip Select and SD Card Interface
--  for Z180 Experimental Computer
--
--  Send SD Card data msb first, not lsb first
--  Supports both byte transfers and bit bashing of SD card data
--  Adopt Martin's idea of having Z180 generate slow clock to shift out data
--
--  Read sd_data (0x48):            Returns last byte transferred from SD card
--  Write sd_data (0x48):           Load transmit shift register. Bit 7 output on MOSI
--  Read sd_exch (0x49):            Returns last byte, then triggers a byte transfer
--  Write sd_exch (0x49):           Loads transmit shift register, then triggers a byte transfer
--  Read sd_ctrl (0x4A or 0x4B):    Bit 0 returns state of MISO, other bits undefined
--                                  Address bit 0 sets the state of sd_pull
--  Write sd_ctrl (0x4A or 0x4B):   Bit 0 sets SD_CS,
--                                  Bit 1 is the bit bash clock (set to zero for byte exchanges)
--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity  Z180_EC_Util_CPLD is
  port (
    data : inout std_logic_vector(7 downto 0);  --  Z180 data bus
    addr : in std_logic_vector(7 downto 0);     --  Z180 port address
    rd, wr, iorq, phi, reset : in std_logic;    --  Z180 control signals
    cs_fpu, rst_fpu, cs_rtc, as_rtc, cs_prop : out std_logic;   -- Chip Selects
    sd_mosi, sd_pull : out std_logic;           --  SD data out
	sd_clk, sd_cs : buffer std_logic;           --  SD control
    sd_miso : in std_logic                      --  SD data return
    );
end Z180_EC_Util_CPLD;

architecture behaviour of Z180_EC_Util_CPLD is
  --  Z180 I/O port addresses
  constant port_rtc_addr : std_logic_vector(7 downto 0) := X"42";   --  RTC address
  constant port_rtc_data : std_logic_vector(7 downto 0) := X"43";   --  RTC data
  constant port_fpu_data : std_logic_vector(7 downto 0) := X"44";   --  FPU data
  constant port_fpu_rst : std_logic_vector(7 downto 0) := X"45";    --  FPU reset
  constant port_sd_data : std_logic_vector(7 downto 0) := X"48";    --  SD read
  constant port_sd_exch : std_logic_vector(7 downto 0) := X"49";    --  SD data
  constant port_sd_ctrl : std_logic_vector(7 downto 0) := X"4A";    --  SD ctrl
  constant port_prop : std_logic_vector(7 downto 0) := X"50";       --  Prop select
  signal sel_sd_data : std_logic;       --  Ports sd_data or sd_exch selected
  signal sel_sd_exch : std_logic;       --  Port sd_exch selected
  signal sel_sd_ctrl : std_logic;       --  Port sd_ctrl selected
  signal sd_arm : std_logic := '0';     --  Trigger SD exchange byte
  signal sd_run : std_logic := '0';     --  Enable SD clock
  signal sd_out : std_logic_vector(7 downto 0); --  Transmit shift register
  signal sd_in : std_logic_vector(8 downto 0);  --  Receive shift register
  signal sd_xclk : std_logic := '0';    --  Bit bash clock
begin
  -- I/O Decoding
  cs_rtc <= iorq when ( addr = port_rtc_data ) else '1';
  as_rtc <= not iorq when ( addr = port_rtc_addr ) and ( wr = '0' ) else '0';
  cs_fpu <= iorq when ( addr = port_fpu_data ) else '1';
  cs_prop <= iorq when ( addr(7 downto 3) = port_prop(7 downto 3) ) else '1';
  -- FPU reset from data(0)
  process (iorq, wr, addr, data)
  begin
    if ( ( iorq = '0' ) and ( wr = '0' ) and ( addr = port_fpu_rst ) ) then
      rst_fpu <= data(0);
    end if;
  end process;
  -- SPI port selects
  sel_sd_data <= iorq when ( addr(7 downto 1) = port_sd_data(7 downto 1) ) else '1';
  sel_sd_exch <= iorq when ( addr = port_sd_exch ) else '1';
  sel_sd_ctrl <= iorq when ( addr(7 downto 1) = port_sd_ctrl(7 downto 1) ) else '1';
  sd_pull <= addr(0) when ( sel_sd_ctrl = '0' ) else '0';
  -- SPI control register
  process (wr, sel_sd_ctrl, data, reset)
  begin
    if ( reset = '0' ) then
      -- Reset to zero
      sd_cs <= '0';
      sd_xclk <= '0';
    elsif ( ( sel_sd_ctrl = '1' )  and ( wr = '0' ) ) then
      -- On write: sd_cs = data(0) and sd_xclk = data(1)
      sd_cs <= data(0);
      sd_xclk <= data(1);
    end if;
  end process;
  -- Clock from PHI, providing sd_run is set
  sd_clk <= ( sd_run and phi ) or sd_xclk;
  -- Arm SPI transfer: After a read or write to port_sd_exch
  process (sel_sd_exch, sd_clk, reset)
  begin
	if ( ( sd_clk = '1' ) or ( reset = '0' ) ) then
      -- Clear sd_arm on reset, or once clock has started
	  sd_arm <= '0';
    elsif ( rising_edge (sel_sd_exch) ) then
      -- Set sd_arm at end of read or write to sd_exch
      sd_arm <= '1';
	end if;
  end process;
  -- Run byte transfer from next low clock
  process (phi, sd_arm, sd_in, reset)
  begin
    if ( reset = '0' ) then
      -- Clear run flag on reset
	  sd_run <= '0';
    elsif ( falling_edge (phi) ) then
      -- Run if sd_arm set, or bits outstanding
      sd_run <= sd_arm or ( not sd_in(8) );
	 end if;
  end process;
  -- Output bits on falling clock edge
  process (sd_clk, sel_sd_data, wr, data)
  begin
    if ( ( sel_sd_data = '0' ) and ( wr = '0' ) ) then
      -- Copy data to output shift register on write to sd_data
      sd_out <= data;
    elsif ( falling_edge (sd_clk) ) then
      -- Shift data out, msb first
      sd_out(7 downto 1) <= sd_out(6 downto 0);
      sd_out(0) <= '1';
    end if;
  end process;
  sd_mosi <= sd_out(7);
  -- Input bits on rising clock edge
  process (sd_clk, sd_arm, sd_run, sd_miso)
  begin
    if ( ( sd_arm = '1' ) and ( sd_run = '0' ) ) then
      -- Transfer will stop when '1' is shifted into high bit.
      sd_in <= "000000001";
    elsif ( rising_edge (sd_clk) ) then
      -- Shift data in
      sd_in(8 downto 1) <= sd_in (7 downto 0);
      sd_in(0) <= sd_miso;
    end if;
  end process;
  -- Read SPI data
  data <= sd_in(7 downto 0) when ( ( rd = '0' ) and ( sel_sd_data = '0' ) ) else
          ( 0 => sd_miso, others => 'Z' ) when ( ( rd = '0' ) and ( sel_sd_ctrl = '0' ) ) else
          ( others => 'Z' );
end behaviour;
